/*
  File: PrismToEMConnector.java

  History:
  Date       	Who				 What
  Jan_31_2004   poladian&jpsousa Refactored *.connector
*/
package connector;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;

import messaging.AbstractPtoPConnector;

import util.Assert;
import util.AuraProperties;
import util.xml.XmlElement;



/**
 * Embodies the Prism end of a Prism-EM connector over TCP. It verifies the alphabet
 * (message tags) but not the protocol (message sequencing).
 * Many client objects can use an instance of this connector concurrently.
 * Sending messages is non-blocking, and conversely the methods called-back upon
 * incoming messages should return as soon as possible.
 * @see edu.cmu.util.AbstractPtoPConnector
 * @see edu.cmu.aura.prism.connector.IPrismCallbacksFromEM
 */
public class PrismToEMConnector extends AbstractPtoPConnector {
	// default network ports and hosts
	private static String __PrismHost = "127.0.0.1";
	private static String __EMhost = "127.0.0.1";
	private static String __PrismListenEMport = "4101";
	private static String __EMlistenPrismPort = "4001";
	// AuraProperties tags for ports and hosts
	private static String __PrismHostProperty =
		"edu.cmu.aura.prism.EMListenHost";
	private static String __EMhostProperty = "edu.cmu.aura.em.PrismListenHost";
	private static String __PrismListenEMportProperty =
		"edu.cmu.aura.prism.EMListenPort";
	private static String __EMlistenPrismPortProperty =
		"edu.cmu.aura.em.PrismListenPort";
	// incoming message tags
	public static final String __CREATEDTASK = "createdTask";
	public static final String __TASKLAYOUT = "taskLayout";
	public static final String __TASK_GONE = "taskGone";
	// outgoing message tags
	public static final String __NEWTASK = "newTask";
	public static final String __BUDGET = "budget";
	public static final String __SETUP = "setup";
	public static final String __DISBAND = "disband";
	// task session identifier attribute
	private static final String __TASKID = "taskId";


	private static final String __SRP_REQUESTOR_LOCATION = "SRPreqLocation";
	private static final String __SRP_REQUESTOR_BIND = "SRPreqBind";

	// member variables
	Vector _waiting = new Vector();
	Hashtable _receivers = new Hashtable();

	/**
	 * Establishes the Prism end of a Prism-EM connector over TCP.  The vales for the TCP endpoints
	 * are read from aura.config or, if missing, hardcoded default values are used.
	 */
	public PrismToEMConnector() {
		// read ports and hosts from AuraProperties
		Properties auraProperties = AuraProperties.getProperties();
		Assert.check(
			auraProperties != null,
			"PrismToEMConnector: Couldn't open AuraProperties");

		__PrismHost =
			auraProperties.getProperty(__PrismHostProperty, __PrismHost);
		__PrismListenEMport =
			auraProperties.getProperty(
				__PrismListenEMportProperty,
				__PrismListenEMport);
		__EMhost = auraProperties.getProperty(__EMhostProperty, __EMhost);
		__EMlistenPrismPort =
			auraProperties.getProperty(
				__EMlistenPrismPortProperty,
				__EMlistenPrismPort);

		try {
			setEndPoints(
				__EMhost,
				(new Integer(__EMlistenPrismPort)).intValue(),
				__PrismHost,
				(new Integer(__PrismListenEMport)).intValue(),
				"PrismToEMConnector");
		} catch (Exception e) {
			Assert.report("PrismToEMConnector: couldn't set the end points");
		}
	}

	/**
	 * Establishes the Prism end of a Prism-EM connector over TCP, between the provided endpoints.
	 * @param eMhost
	 * @param eMport
	 * @param prismHost
	 * @param prismPort
	 */
	public PrismToEMConnector(
		String eMhost,
		String eMport,
		String prismHost,
		String prismPort) {
		try {
			setEndPoints(
				eMhost,
				(new Integer(eMport)).intValue(),
				prismHost,
				(new Integer(prismPort)).intValue(),
				"PrismToEMConnector");
		} catch (Exception e) {
			Assert.report("PrismToEMConnector: couldn't set the end points");
		}
	}

	/**
	 * Queues a newTask message to be sent to the EM.
	 * Each client object wishing to use a Prism-EM connector must initiate a session by
	 * sending a newTask message.
	 * Upon hearing back from the EM, the createdTask method will be called on the client object.
	 * The correspondence between received messages and client objects is automatically established
	 * by the taskId attribute received intially in the createdTask message, and subsequently
	 * in every message incoming from the EM.
	 * Every client object must implement the interface IPrismCallbacksFromEM.
	 * @param client
	 * @param newTaskMessage
	 * @see edu.cmu.aura.prism.connector.IPrismCallbacksFromEM
	 */
	public void sendNewTask(
		IPrismCallbacksFromEM client,
		XmlElement newTaskMessage) {
		addOwnLocationBind(newTaskMessage);
		String tag = newTaskMessage.getName();

		Assert.barf(
			tag.equals(__NEWTASK),
			"PrismToEmConnector: sendNewTask() called with tag " + tag);

		_waiting.add(client);
		send(newTaskMessage.toString());
	}

	/**
	 * Queues a message to be sent to the EM.  Supported messages are all sent by Prism
	 * to the Em, except newTask.
	 * @param message
	 * @see edu.cmu.aura.prism.connector.PrismToEMConnector#sendNewTask(IPrismCallbacksFromEM, XmlElement)
	 */
	public void send(XmlElement message) {
		addOwnLocationBind(message);
		String tag = message.getName();

		Assert.barf(
			tag.equals(__BUDGET)
				|| tag.equals(__SETUP)
				|| tag.equals(__DISBAND),
			"PrismToEmConnector: invalid send tag: " + tag);

		send(message.toString());
	}

	/**
	 * Adds two attributes (__SRP_REQUESTOR_LOCATION and __SRP_REQUESTOR_BIND )
	 * that make up the listening address of this Prism.
	 * Must be called on any send method, using the message that is being sent
	 * @param message
	 * @see edu.cmu.aura.prism.connector.PrismToEMConnector#sendNewTask(IPrismCallbacksFromEM, XmlElement)
	 * @see edu.cmu.aura.prism.connector.PrismToEMConnector#send(XmlElement)
	 */
	protected void addOwnLocationBind(XmlElement message)
	{
		message.setAttribute(__SRP_REQUESTOR_LOCATION, getLocalHost());
		message.setAttribute(__SRP_REQUESTOR_BIND, new String(""+getLocalPort()));
	}

	/* (non-Javadoc)
	 * @see edu.cmu.util.IConnector#errorSending(java.lang.String, java.lang.String, int, java.lang.Exception)
	 */
	public void errorSending(String messageStr, String host, int port, IOException e) {
		try {
			XmlElement message = new XmlElement(messageStr);
			String tag = message.getName();
			String tId = message.getAttribute(__TASKID, null);
			IPrismCallbacksFromEM client;
			if (tag.equals(__NEWTASK))
				client = (IPrismCallbacksFromEM) _waiting.get(0);
			else
				client = (IPrismCallbacksFromEM) _receivers.get(tId);
			client.errorSendingToEM(host, port, e);
		} catch (Exception e1) {
			Assert.report(e1);
		}
	}

	/* (non-Javadoc)
	 * @see edu.cmu.util.IConnector#receive(java.lang.String)
	 */
	public void receive(String messageStr) {
		try {
			XmlElement message = new XmlElement(messageStr);
			String tag = message.getName();
			String tId = message.getAttribute(__TASKID, null);

			if (tag.equals(__CREATEDTASK)) {
				System.out.println("temp debug: " +messageStr);
				IPrismCallbacksFromEM client =
					(IPrismCallbacksFromEM) _waiting.remove(0);
				_receivers.put(tId, client);
				client.createdTask(message);

			} else {
				IPrismCallbacksFromEM client =
					(IPrismCallbacksFromEM) _receivers.get(tId);

				if (tag.equals(__TASKLAYOUT))
					client.taskLayout(message);

				else if (tag.equals(__TASK_GONE)) {
					client.taskGone(message);
					_receivers.remove(tId);
				}
				else
					Assert.report(
						"\n\nPrismToEmConnector got unexpected message:\n\n"
							+ messageStr);
			}
		} catch (Exception e) {
			Assert.report(e);
		}
	}
}
